#include "ClientEventHandler.H"
#include "GameStateModule.H"
#include "GfxModule.H"

#include "GameChanges.H"
#include "GameTile.H"
#include "ServerObjData.H"
#include "Options.H"

#include "EventHandler.H"
#include "Random.H"
#include "Game.H"
#include "Movement.H"
#include "ModuleSet.H"


using namespace std;

enum { TILE_UNKNOWN=0, TILE_WATER=1, TILE_GROUND=2, TILE_CLIFF=3 };


bool ClientEventHandler::handle_event(const Event& e){   

  /* Si el destinatario es GameStateModule */
  if (e.get_who() == GameStateModule::FROM) {

	/* Evento: client dead, game end y/o error */
    if (e.get_what() == GameStateModule::STOP_MSG) {
      state.quit = true;
	  return true;
    }

	/* Evento: socket error  */
    if (e.get_what() == GameStateModule::READ_ERROR_MSG) {
      state.quit = state.error = true;
	  return true;
    }

	/* Evento: view received */
    if (e.get_what() == GameStateModule::VIEW_MSG) {

	  /* Asignamos y mandamos las acciones */
      compute_actions();
	  
	  if(state.quit == true)
	    return true;
		
      state.gsm->send_actions();
	  
	  /* Obtenemos una referencia al juego */
      Game &game = state.gsm->get_game();

	  /* Si modo gfx, se pocesan los cambios */
      if (state.gfxm)
	    state.gfxm->process_changes();

	  /* Obtenemos algunas estadisticas del juego: view frame y action frame */
      sint4 vf = game.get_view_frame();
      sint4 af = game.get_action_frame();

	  /* Si se esta en modo gfx */
      if (state.gfxm) {
        /* No se dibuja si el view frame y el action frame estan muy desfasados */
        if (abs(vf-af) < 10) {
          state.gfxm->draw();
          state.just_drew = true;
        }
      }

      return true;
    }
  }  
   
  return false;
}


void ClientEventHandler::compute_actions(){

	/* Obtenemos una referencia al juego */
	const Game &game = state.gsm->get_game();
	
	/* Obtenemos nuestro id de jugador */
	const sint4 cid = game.get_client_player();
	
	/* Obtenemos un vector de punteros a todos nuestros objetos */
	const Game::ObjCont &objs = game.get_objs(cid);	
	
	/* Obtenemos una referencia al mapa del juego */
	const Map<GameTile> &map = game.get_map();  

	/* Obtenemos una referencia a los cambios que se han producido en el juego */
	const GameChanges &changes = state.gsm->get_changes();
	
	/* Actualizamos el numero de tiles conocidos */
	if (!changes.new_tile_indexes.empty())		
		nTilesKnown += changes.new_tile_indexes.size();		

	/* Obtenemos los objetos iniciales */
	if(!init){
		initGame(objs);
		init = true;
	}
	
	
	if (!changes.dead_objs.empty()){
		
		FORALL(changes.dead_objs, od){			
						
			GameObj* gob = (*od)->get_GameObj();				
			TerrainBase::Loc loc;			
					
			/* Control center propio perdido */
			if((gob) && (*(gob->sod.owner) == cid) && ((*od)->bp_name() == "controlCenter")){				
				gob->get_center(loc.x, loc.y);
				posBase.push_back(getPosBase(loc));				
			
			/* Edificio enemigo destruido */
			}else if ((gob) && (*(gob->sod.owner) != cid) && (((*od)->bp_name() == "controlCenter") || ((*od)->bp_name() == "barracks") || ((*od)->bp_name() == "factory"))){				
				gob->get_center(loc.x, loc.y);				
				
				stringstream aux;
				aux << loc.x;
				aux << loc.y;
				
				string key = aux.str();
				
				if((gob->bp_name() == "controlCenter")){					
					enemyControlCenters.erase(key);
					
				}else{					
					otherEnemyBuildings.erase(key);					
				}
			}			
		}
	}
	
			
	/* Obtenemos el numero de jugadores */
	const sint4 n = game.get_player_num();
	
	/* Se examina el mapa para localizar la base enemiga */
	/* Para cada jugador */
    FORS (i, n) {
		
		/* Si somos nosotros pasamos al siguiente jugador */
		if (i == cid)
			continue;
		
		/* Obtenemos los objetos del jugador i */
		const Game::ObjCont &opp_objs = game.get_objs(i);
		
		/* Para cada uno de los objetos de i */
		FORALL (opp_objs, it2) {				
			
			/* Obtenemos un objeto del jugador i*/
			GameObj * gob2 = (*it2)->get_GameObj();	
			
			/* Si el objeto existe, esta en juego y dentro del mapa (en el servidor), y no esta muerto */
            if (gob2 && gob2->sod.in_game && *gob2->sod.x >= 0 && !gob2->is_dead()) {
				
				ScalarPoint location;
				location.x = *gob2->sod.x;
				location.y = *gob2->sod.y;
				
				stringstream aux;
				aux << location.x;
				aux << location.y;
				
				string key = aux.str();
				
				if(gob2->bp_name() == "factory")
					otherEnemyBuildings[key] = location;
				
				else if(gob2->bp_name() == "barracks")
					otherEnemyBuildings[key] = location;
				
				else if(gob2->bp_name() == "controlCenter")
					enemyControlCenters[key] = location;				
			}				
		}
	}
	
	/* Cuando el numero de control centers cambia hay que recalcular las routas a los minerales */
	if(controlCenters.size() != nControlCenters){		
		nControlCenters = controlCenters.size();		
		for(unsigned i = 0; i < controlCenters.size(); i++)
			simpleGather->computeRoutes(i);
	}
	
	
	/* Comporbamos si algun control center ha sido destruido */
	if(posBase.size() > 0){
		
		/* Solo tenemos una base, el control center ha sido destruido */
		if(bases.size() == 1){			
			
			/* Si disponemos de medios se ordena su construccion */
			if((fm.available() > fm.get_cost("controlCenter")) && ((simpleGather->countWorkers() > 0) || (workers.size() > 0) || (scouting->getNumberScouts() > 0)))
				buildCC = true;
		
		/* Tenemos dos bases y al menos un cc de una ha sido destruido */
		}else if(bases.size() == 2){
		
			/* Han destruido un control center */
			if(posBase.size() == 1){				
			
				sint4 pos = posBase.back();
				
				/* Si esa base no tiene edificios se elimina */
				if(bases[pos]->countBuilding() == 0){										
					
					Vector<GameObj*> units;
					
					/* Se intenta salvar las unidades que existan en su defensa*/
					if((bases[pos]->isDefenderOperative()) && (bases[pos]->countUnitsDefender("") > 0))
						units = bases[pos]->getUnitsToNewBase(3);
									
					Vector<SpecialBaseCommander*>::iterator it = bases.begin();
					it += pos;
					bases.erase(it);
					posBase.pop_back();
					
					FORALL(units, unit)
						bases[0]->addUnitToDefender(*unit);
				}
			
			/* Han destruidos los dos control centers */
			}else{				
			
				/* Hay que eliminar la base con menor numero de edificios */
				sint4 nBuildingB0 = bases[0]->countBuilding();
				sint4 nBuildingB1 = bases[1]->countBuilding();
				sint4 baseToRemove;				
				
				if(nBuildingB0 >= nBuildingB1)
					baseToRemove = 1;		
				else
					baseToRemove = 0;				
				
				Vector<GameObj*> units;
				
				/* Se intenta salvar las unidades que existan en su defensa*/
				if((bases[baseToRemove]->isDefenderOperative()) && (bases[baseToRemove]->countUnitsDefender("") > 0))
					units = bases[baseToRemove]->getUnitsToNewBase(3);
				
				Vector<SpecialBaseCommander*>::iterator it = bases.begin();
				it += baseToRemove;
				bases.erase(it);
				posBase.clear();
				
				FORALL(units, unit)
					bases[0]->addUnitToDefender(*unit);
				
				posBase.push_back(0);
				
				/* Si disponemos de medios se ordena su construccion */
				if((fm.available() > fm.get_cost("controlCenter")) && ((simpleGather->countWorkers() > 0) || (workers.size() > 0) || (scouting->getNumberScouts() > 0)))
					buildCC = true;		
			}
		
		/* Error */
		}else{
			assert(bases.size() <= 2);
		}
	}
	
	
	/* Se construye un nuevo control center */
	if(buildCC){
		buildCC = false;				
		GameObj* unit = NULL;		
		
		/* Obtenemos un worker que estaba realizando gather */
		if(simpleGather->countWorkers() > 0){		
			unit = simpleGather->removeWorkerGather();		
				
		/* Obtenemos un worker que no tenia trabajo asignado*/
		}else if(workers.size() > 0){				
			unit = workers.back();
			workers.pop_back();			
						
		/* Obtenemos un worker que estaba realizando scouting */
		}else if (scouting->getNumberScouts() > 0){			
			unit = scouting->removeScout();						
		}		
		
		/* Se intenta construir si se dispone de una unidad worker */
		if(unit != NULL){
							
			if(fm.available() > fm.get_cost("controlCenter")){
				posBase.clear();
				bases[0]->setAskingForBuilder(true);			
				bases[0]->onReceiveUnit(unit, Watcher::NONE);
				bases[0]->buildBuilding("controlCenter", fm.reserve(fm.get_cost("controlCenter")));				
			}else{				
				workers.push_back(unit);
				unit = NULL;
			}
		}		
	}
	
	
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////       OBTENER VALORES DE ENTRADA       ///////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	/////////////////////////////////////////////
	/////////////// RED DE ATAQUE ///////////////
	/////////////////////////////////////////////
	
	/* Force: Poder militar */	
	size_t nMarines = marines.size();
	size_t nTanks = tanks.size();
		
	FORALL(bases, base){	
		nMarines += (*base)->countUnitsDefender("marine");
		nTanks += (*base)->countUnitsDefender("tank");
	}
		
	float force = nets->getForce(nMarines, nTanks);
	
	
	/* Located enemy?: Indica si se ha localizado alguna base enemiga */
	float localized = 0.0;
	
	if((otherEnemyBuildings.size() > 0) || (enemyControlCenters.size() > 0))
		localized = 1.0;
	
	
	/* Time: Tiempo transcurrido de la partida */
	float finished = nets->getTime(game.get_view_frame());
	
	
	/* Attacking?: Indica si se esta realizando algun ataque */
	float playerAttack = 0.0;
	
	if((attack1 != NULL) || (attack2 != NULL))
		playerAttack = 1.0;
		
	float attackResult = nets->getValueAttack(force, localized, finished, playerAttack);	
	

	/////////////////////////////////////////////
	/////////////// RED DE DEFENSA //////////////
	/////////////////////////////////////////////
	
	Vector<float> defenseResults;
	
	FORALL(bases, base){
		
		/* Enemy attacking?: Indica si el enemigo esta atacando la base */		
		float enemyAttack = 0.0;
	
		if(((*base)->isDefenderOperative()) && ((*base)->isBaseUnderAttack() == true))
			enemyAttack = 1.0;		


		/* Is there building?: Indica si la base cuenta con algun edificio */
		float building = 0.0;
		size_t nBuilding = (*base)->countBuilding();
		
		if(nBuilding > 0)
			building = 1.0;

	
		/* Time last attack */		
		float longTime = nets->getLastAttack((*base)->getLastAttack(), game.get_view_frame());		
		
		float result = nets->getValueDefense(enemyAttack, building, longTime, finished, playerAttack);
		defenseResults.push_back(result);
		
	}
	
	
	//////////////////////////////////////////////
	////////// RED DE CREACION DE BASES //////////
	//////////////////////////////////////////////
	
	/* Minerals: Cantidad de minerales relativa a la construccion de edificios */
	float mineralsBuildingFew = nets->getMineralsBuildingFew(fm.available());
		
	
	/* Have more bases?: Indica si se dispone de mas de una base */
	float yesMoreBases = 0.0;
	
	if(bases.size() > 1)
		yesMoreBases = 1.0;
	
	
	/* In progress?: Indica si se esta llevando a cabo la construnccion de una nueva base */
	float yesInProgress = 0.0;
	
	if(newBase != NULL)
		yesInProgress = 1.0;
		
		
	/* Worker: Indica si se cuenta con al menos una unidad worker */
	float yesWorker = 0.0;
	
	if((simpleGather->countWorkers() > 0) || (workers.size() > 0) || (scouting->getNumberScouts() > 0))
		yesWorker = 1.0;
		
	
	/* Primero obtener los clusters no asignados */	
	Vector<boost::shared_ptr<MineralCluster> > clustersNoAsignados = getClustersNoAsignados(simpleGather->getMineralClusters());
	
	Vector<float> newBaseResults;
	
	FORALL(clustersNoAsignados, cna){
		
		/* Distance old base to cluster minerals */		
		/* Si no hay clusters sin asignar, se considera que la distancia es cero */
		float smallDistanceToMinerals = 1.0;
	
		/* Al menos debe poder haber un cluster sin asignar */
		if(clustersNoAsignados.size() > 0)
			smallDistanceToMinerals = getSmallDistanceToMinerals(*cna);
	
		
		/* Is possible? */		
		float yesPossible = 0.0;
	
		/* Al menos debe poder haber un cluster sin asignar */
		if(clustersNoAsignados.size() > 0)
			yesPossible = isPossible(*cna);
					
		float result = nets->getValueNewBase(mineralsBuildingFew, force, smallDistanceToMinerals, yesMoreBases, yesInProgress, finished, yesPossible, yesWorker);
		newBaseResults.push_back(result);	
	}
	
	/////////////////////////////////////////////
	//// REGLAS DEFENSA + ATAQUE + NEW BASE /////
	/////////////////////////////////////////////
		
	/* Obtenemos el resultado de defensa mas elevado */	
	float maxDefense = -1;
		
	/* Resultados de defense */
	FORALL(defenseResults, result){
		if(*result > maxDefense)
			maxDefense = *result;	
	}
		
	/* Obtenemos el mejor resultado de New Base */	
	float bestBase = -1;
	sint4 posCluster = -1;
	sint4 i = 0;	
	
	/* Resultados de new base */
	FORALL(newBaseResults, result){
		if(*result > bestBase){
			bestBase = *result;
			posCluster = i;
		}	
		i++;
	}
		
		
	/* Vemos si nos estan atacando a alguna base */	
	bool weAreUnderAttack = false;
	
	FORALL(bases, base){	
		if(((*base)->isDefenderOperative()) && ((*base)->isBaseUnderAttack() == true))
			weAreUnderAttack = true;
	}
	
	/* La variable needAttack indica si es viable realizar un ataque (o fundar una base)
	 aunque el valor de defensa sea elevado. Con esto se quiere contrarrestar
	 el efecto de que el enemigo sea muy conservador, es decir, si nunca
	 ataca, el valor de defender va a ser elevado por el tiempo desde el ultimo ataque */
	bool needAction = false;
	
	if((weAreUnderAttack == false) && ((attackResult > 0.7) || (bestBase >= 0.66)))
		needAction = true;	
	
	
	/* Hay que defender */
	if((maxDefense > 0.65) && (needAction == false)){		
				
		/* Dado que todas las unidades siempre van a estar defendiendo 
		a no ser que se este atacando o fundando una nueva base, cuando 
		se necesite defender se va a estudiar la posibilidad de cancelar
		 un ataque y utilizar esas unidades en la defensa de la base o 
		 bases afectadas. Nunva se va a cancelar el proceso de creacion 
		 de una nueva base */
		
		if(attack1 != NULL)
			evaluateAttack(&attack1, weAreUnderAttack);
			
		if(attack2 != NULL)
			evaluateAttack(&attack2, weAreUnderAttack);
	
	/* Hay que fundar una nueva base */
	}else if((bestBase >= 0.66) && (bases.size() == 1) && (bases[0]->isDefenderOperative()) && ((otherEnemyBuildings.size() > 0) || (enemyControlCenters.size() > 0))){
				
		if(fm.available() > fm.get_cost("controlCenter")){
			createNewBase(clustersNoAsignados[posCluster]);
			mineralsBuildingFew = nets->getMineralsBuildingFew(fm.available());
		}		
	
	/* Hay que atacar */
	}else if((attackResult > 0.7) && ((attack1 == NULL) || (attack2 == NULL))){
		
		if((needAction == true) && (maxDefense > 0.65)){			
			createAttack("low");
		
		}else if(attackResult > 0.95){			
			createAttack("high");
			
		}else{			
			createAttack("medium");
		}
		
		/* Actualizamos entradas a la redes posteriores (force y playerAttack)*/
		size_t nMarines = marines.size();
		size_t nTanks = tanks.size();
		
		FORALL(bases, base){	
			nMarines += (*base)->countUnitsDefender("marine");
			nTanks += (*base)->countUnitsDefender("tank");
		}
		
		force = nets->getForce(nMarines, nTanks);
			
		playerAttack = 1.0;		
			
	}	
	

	/* Las unidades militares sin tareas asignadas,
	 se integran en la defensa de la base mas cercana */
	defaultDefense(&marines);
	defaultDefense(&tanks);	


	/////////////////////////////////////////////
	////////////////// BUILDING /////////////////
	/////////////////////////////////////////////
	
	/* Calculamos todas las entradas necesarias para las redes de construccion de Barracks y Factory */
	/* Primero se calculan las entradas independientes de las bases */
	
	/* BARRACKS */
		
	/* Is there control center?: Indica si al menos se cuenta con un edificio control center */	
	float yesControlCenter = 0.0;
	
	FORALL(bases, base){
		if((*base)->isThereBuilding("controlCenter"))
			yesControlCenter = 1.0;
	}	
		
	Vector<float> barracksResults;
	
	FORALL(bases, base){
				
		/* Is there barracks?: Indica si la base actual dispone del edifico barracks */		
		float yesBarracks = (*base)->isThereBuilding("barracks");
				
	
		/* Building now?: Indica si se esta construyendo el edificio barracks (en la base actual) */		
		float buildingNowBarracks = (*base)->buildingNow("barracks");
						
		
		/* Distance base to enemy base: Distancia de la base (actual) al edificio enemigo mas cercano */		
		Vector<ScalarPoint> locationsEnemyBuilding;
	
		for(it = otherEnemyBuildings.begin(); it != otherEnemyBuildings.end(); it++)		
			locationsEnemyBuilding.push_back(it->second);
	
		for (it = enemyControlCenters.begin(); it != enemyControlCenters.end(); it++)
			locationsEnemyBuilding.push_back(it->second);
	
		float smallDistanceToEnemyBase = nets->getSmallDistanceToEnemyBase((*base)->getCentre(), locationsEnemyBuilding);
		
				
		float result = nets->getValueBuildingBarracks(mineralsBuildingFew, yesBarracks, buildingNowBarracks, finished, yesMoreBases, smallDistanceToEnemyBase, yesControlCenter, yesWorker);
		barracksResults.push_back(result);		
			
	}
				
	
	/* Is there barracks?: Indica si al menos se cuenta con un edificio barracks */	
	float yesBarracks2 = 0.0;
	
	FORALL(bases, base){
		if((*base)->isThereBuilding("barracks"))
			yesBarracks2 = 1.0;
	}
		
	
	Vector<float> factoryResults;
	
	
	FORALL(bases, base){	
	
		/* Is there factory?: Indica si la base actual cuenta con un edificio factory */
		float yesFactory = (*base)->isThereBuilding("factory");
	
	
		/* Building now?: Indica si en la base actual se esta construyendo el edificio factory */		
		float buildingNowFactory = (*base)->buildingNow("factory");
			
		float result = nets->getValueBuildingFactory(mineralsBuildingFew, yesFactory, buildingNowFactory, finished, yesBarracks2, yesControlCenter, yesWorker);
		factoryResults.push_back(result);		
	}	
	
	
	/* Se analiza y procesa los resultados de las redes building */	
	/* No es recomendable realizar dos construcciones simultaneas, ya que es una inversion muy elevada */
		
	sint4 posBarracks = -1;
	float maxBarracks = -1;
	i = 0;
	
	/* Resultados de barracks */
	FORALL(barracksResults, result){
		if(*result > maxBarracks){
			maxBarracks = *result;
			posBarracks = i;
		}	
		i++;
	}

	sint4 posFactory = -1;
	float maxFactory = -1;
	i = 0;
		
	/* Resultados de factory */
	FORALL(factoryResults, result){
		if(*result > maxFactory){
			maxFactory = *result;
			posFactory = i;
		}	
		i++;
	}


	/* Se comprueba si se debe realizar alguna construccion */
	if(((maxFactory > 0.50) || (maxBarracks > 0.50)) && ((bestBase < 0.66) || ((bestBase >= 0.66) && ((maxFactory > 0.68) || (maxBarracks > 0.68))))){
		
		GameObj* unit = NULL;

			
		/* Obtenemos un worker que estaba realizando gather */
		if(simpleGather->countWorkers() > 0){			
			unit = simpleGather->removeWorkerGather();			
			
		/* Obtenemos un worker que no tenia trabajo asignado*/
		}else if(workers.size() > 0){				
			unit = workers.back();
			workers.pop_back();			
		
		/* Obtenemos un worker que estaba realizando scouting */
		}else if (scouting->getNumberScouts() > 0){			
			unit = scouting->removeScout();			
		}

		
		/* Se intenta construir si se dispone de una unidad worker */
		if(unit != NULL){
	
			/* Se debe construir un edificio barracks */
			if(maxBarracks > maxFactory){				
								
				if(fm.available() > fm.get_cost("barracks")){
					bases[posBarracks]->setAskingForBuilder(true);			
					bases[posBarracks]->onReceiveUnit(unit, Watcher::NONE);
					bases[posBarracks]->buildBuilding("barracks", fm.reserve(fm.get_cost("barracks")));				
				}else{					
					workers.push_back(unit);
					unit = NULL;
				}
		
			/* Se debe construir un edificio factory */
			}else{		
								
				if(fm.available() > fm.get_cost("factory")){
					bases[posFactory]->setAskingForBuilder(true);			
					bases[posFactory]->onReceiveUnit(unit, Watcher::NONE);
					bases[posFactory]->buildBuilding("factory", fm.reserve(fm.get_cost("factory")));				
				}else{					
					workers.push_back(unit);
					unit = NULL;
				}
			}		
		}			
	}
	

	/////////////////////////////////////////////
	////////////////// TRAINING /////////////////
	/////////////////////////////////////////////	
	
	
	/* Minerals: Cantidad de minerales relativa al entrenamiento de unidades */
	float mineralsTrainingFew = nets->getMineralsTrainingFew(fm.available());
	
	Vector<float> workersResults;
	
	FORALL(bases, base){
	
		/* Can gather?: Indica se hay vacantes para realizar la tarea gather en la base */
		bool yesCanGather = simpleGather->canGather((*base)->clusters, ((*base)->countQueued("worker") + workers.size()));
		
		
		/* Queue: Indica si la cola de entrenamiento esta llena */		
		size_t workersQueued = (*base)->countQueued("worker");
		float full = 0.0;
		
		if(workersQueued == 6)
			full = 1.0;
					
						
		float result =  nets->getValueTrainingWorker(mineralsTrainingFew, finished, yesControlCenter, yesCanGather, full);
		workersResults.push_back(result);	
		
	}
	
	sint4 posWorker = 0;
	
	/* Se estudian los resultados, si alguno es superior al 80% se entrena */
	FORALL(workersResults, result){
	
		/* Se debe entrenar una unidad worker */
		if(*result > 0.80){
								
			if (bases[posWorker]->canBuild("worker") && fm.available() > fm.get_cost("worker")){			
				bases[posWorker]->buildUnit("worker", fm.reserve(fm.get_cost("worker")));
				mineralsTrainingFew = nets->getMineralsTrainingFew(fm.available());				
			}			
		}
		
		posWorker++;
	}
	
	/* Unidades militares */

	/* Number of units: Representa la cantidad de unidades del mismo tipo */
	/* Como esta red se usa tanto para marines como para tanks,
	 hay que calcular esta entrada para ambos casos */
	size_t countMarines = marines.size();
	size_t countTanks = tanks.size();
	
	FORALL(bases, base){
		countMarines += (*base)->countQueued("marine");
		countMarines += (*base)->countUnitsDefender("marine");
		countTanks += (*base)->countQueued("tank");
		countTanks += (*base)->countUnitsDefender("tank");
	}
		
	float minMarine = nets->getMinNumberOfUnit("marine", countMarines);
	float minTank = nets->getMinNumberOfUnit("tank", countTanks);
	
	
	/* Rate: Indica el ratio actual entre marines y tanks */
	/* Como esta red se usa tanto para marines como para tanks,
	 hay que calcular esta entrada para ambos casos */	
	float koMarine = nets->getKoRatio("marine", countMarines, countTanks);
	float koTank = nets->getKoRatio("tank", countMarines, countTanks);
	

	Vector<float> marinesResults;
	Vector<float> tanksResults;
		
	FORALL(bases, base){	
	
		/* Is there building? */
		/* Como esta red se usa tanto para marines como para tanks,
		 hay que calcular esta entrada para ambos casos */		
		bool yesBuildingBarracks = (*base)->isThereBuilding("barracks");
		bool yesBuildingFactory = (*base)->isThereBuilding("factory");
	
	
		/* Queue: Indica si hay hueco en las colas de entrenamiento */
		/* Como esta red se usa tanto para marines como para tanks,
		 hay que calcular esta entrada para ambos casos */
		size_t marinesQueued = (*base)->countQueued("marine");
		size_t tansQueued = (*base)->countQueued("tank");
		size_t nBarracks = (*base)->countBuildings("barracks");
		size_t nFactory = (*base)->countBuildings("factory");
		
		float noFullMarines = 0.0;
		float noFullTanks = 0.0;
		
		if(marinesQueued < (6 * nBarracks))
			noFullMarines = 1.0;
		
		if(tansQueued < (6 * nFactory))
			noFullTanks = 1.0;
					
			
		float result = nets->getValueTrainingUnit(minMarine, koMarine, mineralsTrainingFew, finished, yesBuildingBarracks, noFullMarines);
		marinesResults.push_back(result);
				
		result = nets->getValueTrainingUnit(minTank, koTank, mineralsTrainingFew, finished, yesBuildingFactory, noFullTanks);
		tanksResults.push_back(result);
	}
	
	/* Ahora se estudian los resultados y si alguno es superior al 40% se entrena la unidad */
	sint4 posUnit = 0;
	float maxBuilding = max(maxBarracks, maxFactory);

	FORALL(marinesResults, result){
		
		/* Se debe entrenar una unidad marine y/o una unidad tank si se debe y no se debe realizar una construccion urgente o si se debe pero se cuenta con poco poder militar */
		//if(((marinesResults[posUnit] > 0.40) || (tanksResults[posUnit] > 0.40)) && ((maxBuilding < 0.65) || ((maxBuilding > 0.65) && (force < (0.05))))){
		if(((marinesResults[posUnit] > 0.40) || (tanksResults[posUnit] > 0.40)) && ((maxBuilding < 0.65) || ((maxBuilding > 0.65) && (force < (0.05)))) && ((bestBase < 0.66) || ((bestBase >= 0.66) && (force < (0.4))))                    ){
									
			/* Si la diferencia de los resultados es notable solo le entrena la unidad con valor mas alto */
			if((marinesResults[posUnit] > 0.40) && (tanksResults[posUnit] > 0.40) && (abs(tanksResults[posUnit] - marinesResults[posUnit]) >= 0.14)){
				
				/* Se entrena una unidad marine */
				if(marinesResults[posUnit] > tanksResults[posUnit]){					
										
					if (bases[posUnit]->canBuild("marine") && fm.available() > fm.get_cost("marine"))
						bases[posUnit]->buildUnit("marine", fm.reserve(fm.get_cost("marine")));					
					
				/* Se entrena una unidad tank */
				}else{					
				
					if (bases[posUnit]->canBuild("tank") && fm.available() > fm.get_cost("tank"))
						bases[posUnit]->buildUnit("tank", fm.reserve(fm.get_cost("tank")));					
				}
			
			/* Se entrena las unidades que se deban */
			}else{
				
				/* Se entrena una unidad marine */
				if(marinesResults[posUnit] > 0.40){					
				
					if (bases[posUnit]->canBuild("marine") && fm.available() > fm.get_cost("marine"))
						bases[posUnit]->buildUnit("marine", fm.reserve(fm.get_cost("marine")));				
				}
				
				
				/* Se entrena una unidad tank */
				if(tanksResults[posUnit] > 0.40){					
				
					if (bases[posUnit]->canBuild("tank") && fm.available() > fm.get_cost("tank"))
						bases[posUnit]->buildUnit("tank", fm.reserve(fm.get_cost("tank")));					
				}
			}			
		}		
		posUnit++;
	}
		
			
//PROCEDIMIENTO PARA CREAR UNA NUEVA BASE
				
	
	if(newBase != NULL){		
		
		bool success = newBase->run();	
		
		/* Todo esta listo para construir la nueva base */
		if(newBase->isFinish()){
			buildNewBase(map);	
		}		
		
		/* Hay problemas */
		if(success == false){
			examineProblems();
		}		
	}
	

	//PROCEDIMIENTO PARA EL ATAQUE 1

	if(attack1 != NULL){
		attack1->update();
		
		/* Se ha terminado el ataque con exito */
		if((attack1->isUnderAttack() == false) && (attack1->isUnderControl()) && (attack1->countUnits("") > 0)){
						
			Vector<GameObj*> units = attack1->getAllUnits();
			
			assert(units.size() > 0);
			
			delete attack1;
			attack1 = NULL;
			
			GameObj* gob = units.back();
			TerrainBase::Loc locUnits;			
			gob->get_center(locUnits.x, locUnits.y);
			
			TerrainBase::Loc locEnemy;
			locEnemy = getLocationEnemyToNewBase(locUnits);
			
			attack1 = new SpecialDefenseCommander(ms, locEnemy, 128, 0, false);
			FORALL(units, unit)
				attack1->addUnit(*unit);
		
		/* El ataque se ha terminado sin exito */
		}else if(attack1->countUnits("") <= 0){			
			delete attack1;
			attack1 = NULL;
		}
	}
	
		
	//PROCEDIMIENTO PARA EL ATAQUE 2
	
	if(attack2 != NULL){
		attack2->update();
		
		/* Se ha terminado el ataque con exito */
		if((attack2->isUnderAttack() == false) && (attack2->isUnderControl()) && (attack2->countUnits("") > 0)){			
			
			Vector<GameObj*> units = attack2->getAllUnits();
			
			assert(units.size() > 0);
			
			delete attack2;
			attack2 = NULL;
			
			GameObj* gob = units.back();
			TerrainBase::Loc locUnits;			
			gob->get_center(locUnits.x, locUnits.y);
			
			TerrainBase::Loc locEnemy;
			locEnemy = getLocationEnemyToNewBase(locUnits);
			
			attack2 = new SpecialDefenseCommander(ms, locEnemy, 128, 0, false);
			FORALL(units, unit)
				attack2->addUnit(*unit);
			
		/* El ataque se ha terminado sin exito */
		}else if(attack2->countUnits("") <= 0){			
			delete attack2;
			attack2 = NULL;
		}
	}
		 
	
	/////////////////////////////////////////////
	////////////////// SCOUTING /////////////////
	/////////////////////////////////////////////
		
	
	size_t nEnemyBuilding = enemyControlCenters.size() + otherEnemyBuildings.size();

	/* Se realiza scouting cuando no se ha descubierto ningun edificio enemigo o cuando
	 se conoce menos del 70% del mapa y se tiene localizado al menos un edifico enemigo */
	if((((nTilesKnown/40.96) < 70) && (nEnemyBuilding > 0)) || (nEnemyBuilding == 0)){

		/* Se añaden nuevos scouts */
		if(scouting->getNumberScouts() < NUMBER_OF_SCOUTS){

			for(unsigned i = scouting->getNumberScouts(); i < NUMBER_OF_SCOUTS; i++){

				/* Utilizamos un worker sin tarea asignada */
				if(workers.size() > 0){
					GameObj * gob = workers.back();
					workers.pop_back();
					scouting->addScout(gob);

				/* Utilizamos un worker que este recolectando minerales */
				}else if(simpleGather->countWorkers() > 0){
					GameObj * gob = simpleGather->removeWorkerGather();
					scouting->addScout(gob);				
				}							
			}
		}

		/* Se realiza la tarea scout si se dispone de al menos un scout */
		if(scouting->getNumberScouts() > 0)
			scouting->scouting(map);

	}else{	

		/* Liberamos a los scouts */
		while(scouting->getNumberScouts() > 0){
			GameObj * gob = scouting->removeScout();
			workers.push_back(gob);			
		}		
	}
	
	
	/////////////////////////////////////////////
	/////////////////// GATHER //////////////////
	/////////////////////////////////////////////
	

	/* Miramos si hemos encontrado nuevos minerales o alguno se ha terminado */		
	if(!(simpleGather->isEmptyNewMinerals()) || (simpleGather->getMineralDied())){
				
		simpleGather->administerMinerals();
		
		/* Siempre que hay un cambio en los minerales hay que asignar 
		 los clusters a las bases y los minerales a los workers */		
		if(simpleGather->newChanges()){
			
			/* Obtenemos los clusters */
			Vector<boost::shared_ptr<MineralCluster> > clusters = simpleGather->getMineralClusters();
			
			/* Tenemos que esperar nuevos cambios */
			simpleGather->setChanges(false);
			
			/* Un cluster pertenece a una base si dista menos de 200 puntos
			 del control center y si el cluster es accesible */
			asignarClusters(clusters, map);
			
			/* Comprobamos si los clusters asignados a los workers siguen siendo validos */
			simpleGather->comprobarWorkersGather();			
		}
	}	

	/* Si necesitamos recolectar y hay clusters disponibles */
	if((needGather) && (simpleGather->getNumberMineralClusters() > 0)){

		/* Si no se han formado los grupos o hay workers libres */
		if((simpleGather->getNumberWorkersGather() == 0) || !workers.empty()){
			
			size_t j = workers.size();
			
			/* Se intentan usar todas las unidades disponibles */
			for(size_t i = 0; i < j; i++){
			
				bool success = false;
				
				/* Examinamos todas las bases */
				FORALL(bases, base){
							
					/* Comprobamos si en la base actual hay hueco para mas unidades */
					if((simpleGather->canGather((*base)->clusters, 0)) && (workers.size() > 0)){
				   
						/* obtenemos el control center de la base */
						GameObj* ccToGather = (*base)->getControlCenter();
					
						/* Si no hay cc no se asignan workers */
						if(ccToGather != NULL){
				   
							/* Obtenemos los clusters asignados a la base */
							Vector<boost::shared_ptr<MineralCluster> > clustersAsignados = (*base)->clusters;
							
							/* Si no hay asignado ningun cluster, no se puede realizar gather */
							if(clustersAsignados.size() > 0){
								
								/* obtenemos el mejor cluster */
								boost::shared_ptr<MineralCluster> bestCluster = simpleGather->getBestCluster(clustersAsignados);
				   
								assert(bestCluster != NULL);				   
				   
								/* obtenemos la unidad worker */
								GameObj* gob = workers.back();
								workers.pop_back();
				   
								assert(gob != NULL);
														
								if(success)
									i++;
				   
								simpleGather->addWorkersGather(gob, ccToGather, bestCluster);							
								success = true;
							}
						}
					}
				}
			}
		}		
		simpleGather->gather();		
	
	}	
		
	Vector<SpecialBaseCommander*>::iterator it;
				
	/* Actualizamos todas las bases */
	for(it = bases.begin(); it != bases.end(); it++)	
		(*it)->update();
}


/* Obtenemos la informacion inicial e inicializamos determinados elementos */
void ClientEventHandler::initGame(const Game::ObjCont &objs){	
   			
	nets->init();
	Vector<GameObj*> units;
	GameObj* cc;
	
	/* Para cada objeto nuestro */
	FORALL (objs, it) {
		
		/* Obtenemos un worker */
		GameObj * gob = (*it)->get_GameObj();
		
		if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
		
			if(gob->bp_name() == "worker"){				
				units.push_back(gob);
			
			}else if(gob->bp_name() == "controlCenter"){				
				units.push_back(gob);
				cc = gob;				
			}
		}
	}
	
	TerrainBase::Loc loc;
	cc->get_center(loc.x, loc.y);
	
	SpecialBaseCommander *bc = new SpecialBaseCommander(ms, units, loc, &controlCenters, &workers, &marines, &tanks);
	bases.push_back(bc);
}


void ClientEventHandler::asignarClusters(Vector<boost::shared_ptr<MineralCluster> > clusters, const Map<GameTile> &map){

	TerrainBase::Loc locCC;
	TerrainBase::Loc locCluster;	
	TerrainBase::Loc tileCC, tileCCParalela1, tileCCParalela2;	
	TerrainBase::Loc tileCluster, tileClusterParalela1, tileClusterParalela2;
	float distance;

	/* A cada base se le asigna los clusters que le pertenezcan */
	FORALL(bases, base){
		
		/* Obtenemos la posición del control center de la base */
		locCC = (*base)->getLocationControlCenter();
				
		/* Vaciamos el vector de clusters asociados a la base */
		(*base)->clusters.clear();
				
		/* Se comprueba que la base disponga de control center */
		if((locCC.x != -1) && (locCC.y != -1)){
					
			/* Calculamos el tile del Control Center */
			tileCC = getTile(locCC);			
						
			/* Se evaluan todos los clusters */
			FORALL(clusters, cluster){

				/* Se obtiene el centro del cluster */
				locCluster = (*cluster)->getCentroid();
								
				/* Se calcula la distancia del CC al centro del cluster */
				distance = sqrt(square(locCC.x - locCluster.x) + square(locCC.y - locCluster.y));				
								
				/* Calculamos el tile del cluster */
				tileCluster = getTile(locCluster);
								
				
				/* Se comprueba que la distancia sea como maximo de 200 puntos */
				if(distance <= 200){
				
					/* Ahora se comprueba que el cluster sea accesible desde el control center.
					 Para ello se calcula la ruta en linea recta, una ruta paralela superior y
					 una ruta paralela inferior. Si alguna de ellas es accesible, el cluster se
					 asigna a la base. La rutas paralelas son las mas cercanas a la central. */
										
					/* El cluster y el control center estan alineados verticalmente */
					if((tileCC.x == tileCluster.x) && (tileCC.y != tileCluster.y)){
										
						tileCCParalela1.x = tileCC.x +1;
						tileCCParalela1.y =	tileCC.y;
						tileClusterParalela1.x = tileCluster.x +1;
						tileClusterParalela1.y = tileCluster.y;
						
						tileCCParalela2.x = tileCC.x -1;
						tileCCParalela2.y = tileCC.y;
						tileClusterParalela2.x = tileCluster.x -1;
						tileClusterParalela2.y = tileCluster.y;
															
					/* El cluster y el control center estan alineados horizontalmente */
					}else if((tileCC.x != tileCluster.x) && (tileCC.y == tileCluster.y)){
																	
						tileCCParalela1.x = tileCC.x;
						tileCCParalela1.y =	tileCC.y +1;
						tileClusterParalela1.x = tileCluster.x;
						tileClusterParalela1.y = tileCluster.y +1;
						
						tileCCParalela2.x = tileCC.x;
						tileCCParalela2.y = tileCC.y -1;
						tileClusterParalela2.x = tileCluster.x;
						tileClusterParalela2.y = tileCluster.y -1;
					
					/* Arriba a la izquierda, el cluster respecto al control center */
					}else if((tileCC.x != tileCluster.x) && (tileCC.y != tileCluster.y) && (tileCC.x > tileCluster.x) && (tileCC.y > tileCluster.y)){
											
						tileCCParalela1.x = tileCC.x;
						tileCCParalela1.y =	tileCC.y -1;
						tileClusterParalela1.x = tileCluster.x +1;
						tileClusterParalela1.y = tileCluster.y;
						
						tileCCParalela2.x = tileCC.x -1;
						tileCCParalela2.y = tileCC.y;
						tileClusterParalela2.x = tileCluster.x;
						tileClusterParalela2.y = tileCluster.y +1;
					
					/* Abajo a la izquierda, el cluster respecto al control center */
					}else if((tileCC.x != tileCluster.x) && (tileCC.y != tileCluster.y) && (tileCC.x > tileCluster.x) && (tileCC.y < tileCluster.y)){
											
						tileCCParalela1.x = tileCC.x -1;
						tileCCParalela1.y =	tileCC.y;
						tileClusterParalela1.x = tileCluster.x;
						tileClusterParalela1.y = tileCluster.y -1;
						
						tileCCParalela2.x = tileCC.x;
						tileCCParalela2.y = tileCC.y +1;
						tileClusterParalela2.x = tileCluster.x +1;
						tileClusterParalela2.y = tileCluster.y;
					
					/* Arriba a la derecha, el cluster respecto al control center */
					}else if((tileCC.x != tileCluster.x) && (tileCC.y != tileCluster.y) && (tileCC.x < tileCluster.x) && (tileCC.y > tileCluster.y)){
											
						tileCCParalela1.x = tileCC.x +1;
						tileCCParalela1.y =	tileCC.y;
						tileClusterParalela1.x = tileCluster.x;
						tileClusterParalela1.y = tileCluster.y +1;
						
						tileCCParalela2.x = tileCC.x;
						tileCCParalela2.y = tileCC.y -1;
						tileClusterParalela2.x = tileCluster.x -1;
						tileClusterParalela2.y = tileCluster.y;
					
					/* Abajo a la derecha, el cluster respecto al control center */
					}else if((tileCC.x != tileCluster.x) && (tileCC.y != tileCluster.y) && (tileCC.x < tileCluster.x) && (tileCC.y < tileCluster.y)){
											
						tileCCParalela1.x = tileCC.x +1;
						tileCCParalela1.y =	tileCC.y;
						tileClusterParalela1.x = tileCluster.x;
						tileClusterParalela1.y = tileCluster.y -1;
						
						tileCCParalela2.x = tileCC.x;
						tileCCParalela2.y = tileCC.y +1;
						tileClusterParalela2.x = tileCluster.x -1;
						tileClusterParalela2.y = tileCluster.y;					
					
					}										
					
					if((rutaAccesible(tileCC, tileCluster, map)) || (rutaAccesible(tileCCParalela1, tileClusterParalela1, map)) || (rutaAccesible(tileCCParalela2, tileClusterParalela1, map)))
						(*base)->clusters.push_back(*cluster);				
				}				
			}		
		}		
	}
}


bool ClientEventHandler::rutaAccesible(TerrainBase::Loc locA, TerrainBase::Loc locB, const Map<GameTile> &map){

	bool success = true;
	int dx, dy;
	int pasos;
	float xInc, yInc;
	float x, y;
	
	/* Se obtienen los puntos intermedios entre A y B,
	 utilizando el algoritmo Diferencial Digital */
	dx = locB.x - locA.x;
	dy = locB.y - locA.y;
	
	if(abs(dx) > abs(dy))
		pasos = abs(dx);
	else
		pasos = abs(dy);
		
	xInc = (float) dx / pasos;
	yInc = (float) dy / pasos;
		
	x = locA.x;
	y = locA.y;			
	
	const GameTile &tile = map(round(x), round(y));
	
	if((tile.is_split()) || (tile.get_typeW() == TILE_UNKNOWN) || (tile.get_typeW() == TILE_CLIFF))
		success = false;
			
	if(success){
		for(int k = 1; k <= pasos; k++){
			x += xInc;
			y += yInc;			
				
			const GameTile &tile2 = map(round(x), round(y));
			
			if((tile2.is_split()) || (tile2.get_typeW() == TILE_UNKNOWN) || (tile2.get_typeW() == TILE_CLIFF)){
				success = false;
				k = pasos + 1;
			}		
		}
	}	
	
	return success;
}


TerrainBase::Loc ClientEventHandler::getTile(TerrainBase::Loc point){

	TerrainBase::Loc tile;
		
	tile.x = point.x / 16;	
	tile.y = point.y / 16;	
	
	assert(point.x < 1024);
	assert(point.y >= 0);
	
	assert(point.y < 1024);
	assert(point.x >= 0);
	
	assert(tile.x < 64);
	assert(tile.x >= 0);
	
	assert(tile.x < 64);
	assert(tile.y >= 0);
	
	return tile;
}


Vector<boost::shared_ptr<MineralCluster> > ClientEventHandler::getClustersNoAsignados(Vector<boost::shared_ptr<MineralCluster> > mineralClusters){
	
	Vector<boost::shared_ptr<MineralCluster> > clustersNoAsignados;
	
	/* Se copian todos los clusters del escenario en el vector de
	 los no asignados y posteriormente se eliminan lo que si han
	 sido asignados a alguna base */
	FORALL(mineralClusters, cluster){
		clustersNoAsignados.push_back(*cluster);
	}	
	
	/* Se busca en todas las bases */
	FORALL(bases, base){	
	
		/* Se eliminan todos los clusters asignados a la base  */
		FORALL((*base)->clusters, cb){
						
			int j = clustersNoAsignados.size();
			Vector<boost::shared_ptr<MineralCluster> >::iterator it = clustersNoAsignados.begin();
			
			/* Se busca el cluster en todo el vector */
			for(int i = 0; i < j; i++){
				
				if((*it) == (*cb)){
					clustersNoAsignados.erase(it);
					i = j;
				}else{
					it++;
				}
			}		
		}
	}	
	
	return clustersNoAsignados;
}


float ClientEventHandler::getSmallDistanceToMinerals(boost::shared_ptr<MineralCluster> cluster){

	float big;
	float small;
	float distance = 0;
	TerrainBase::Loc locCluster = cluster->getCentroid();	
	
	FORALL(bases, base){		
		TerrainBase::Loc locCC = (*base)->getLocationControlCenter();		
		distance += sqrt(square(locCC.x - locCluster.x) + square(locCC.y - locCluster.y));		
	}
	
	distance = (float)distance / bases.size();
		
	big = (float)distance / 400;
	
	if(big > 1.0)
		big = 1.0;
		
	small = 1.0 - big;
		
	return small;
}


float ClientEventHandler::isPossible(boost::shared_ptr<MineralCluster> cluster){

	float possible = 1.0;
	float distance;
	TerrainBase::Loc locCluster = cluster->getCentroid();

	/* Si no se ha localizado al enemigo no es posible crear
	 una nueva base, ya que es muy peligroso */
	if((otherEnemyBuildings.size() > 0) || (enemyControlCenters.size() > 0)){
	
		FORALL(otherEnemyBuildings, building){
			ScalarPoint loc = building->second;
			distance = sqrt(square(loc.x - locCluster.x) + square(loc.y - locCluster.y));
			if(distance < 140){
				possible = 0.0;				
			}
		}
		
		if(possible){
			FORALL(enemyControlCenters, building){
				ScalarPoint loc = building->second;
				distance = sqrt(square(loc.x - locCluster.x) + square(loc.y - locCluster.y));
				if(distance < 140){
					possible = 0.0;					
				}
			}
		}
	
	}else{
		possible = 0.0;		
	}
	
	return possible;
}


void ClientEventHandler::defaultDefense(Vector<GameObj*> * vector){
	
	float distance;
	Vector<GameObj*> unitsToRemove;
	
	/* Se inserta cada unidad en la defensa de la base mas cercana */		
	FORALL((*vector), unit){
		
		sint4 bestDistance = 1448;
		sint4 nBase = -1;
		sint4 i = 0;
		TerrainBase::Loc locUnit;
		(*unit)->get_center(locUnit.x, locUnit.y);
		
		FORALL(bases, base){
			
			TerrainBase::Loc locBase = (*base)->getCentre();		
			distance += sqrt(square(locBase.x - locUnit.x) + square(locBase.y - locUnit.y));
			
			if(distance < bestDistance){
				bestDistance = distance;
				nBase = i;
			}
			i++;
		}
		
		if((nBase != -1) && (bases[nBase]->isDefenderOperative())){
			bases[nBase]->addUnitToDefender(*unit);
			unitsToRemove.push_back(*unit);			
		}		
	}
	
	FORALL(unitsToRemove, gob){
		Vector<GameObj*>::iterator it = vector->find(*gob);
		
		if(it != vector->end()){
			vector->erase(it);			
		}			
	}	
	unitsToRemove.clear();	
}


void ClientEventHandler::createNewBase(boost::shared_ptr<MineralCluster> cluster){	
	
	/* Obtenenmos un tercio del poder militar actual */
	Vector<GameObj*> vectorUnidades = bases[0]->getUnitsToNewBase(1);
		
	GameObj* unit = NULL;
			
	/* Obtenemos un worker que no tenia trabajo asignado*/
	if(workers.size() > 0){		
		unit = workers.back();
		workers.pop_back();		
		
	/* Obtenemos un worker que estaba realizando gather */
	}else if(simpleGather->countWorkers() > 0){		
		unit = simpleGather->removeWorkerGather();		
		
	/* Obtenemos un worker que estaba realizando scouting */
	}else if (scouting->getNumberScouts() > 0){		
		unit = scouting->removeScout();		
	}
			
	assert(unit != NULL);
	
	/* Obtenemos la posicion del edificio enemigo mas cercano al cluster */
	TerrainBase::Loc locEnemy = getLocationEnemyToNewBase(cluster->getCentroid());	
	
	newBase = new NewBase(ms, vectorUnidades, unit, cluster->getCentroid(), locEnemy, fm.reserve(fm.get_cost("controlCenter")));	
}


TerrainBase::Loc ClientEventHandler::getLocationEnemyToNewBase(TerrainBase::Loc locCluster){

	float bestDistance = 1449;
	float distance;
	TerrainBase::Loc enemyLoc;
	
	FORALL(otherEnemyBuildings, building){
		ScalarPoint loc = building->second;
		distance = sqrt(square(loc.x - locCluster.x) + square(loc.y - locCluster.y));
		if(distance < bestDistance){
			bestDistance = distance;
			enemyLoc.x = loc.x;
			enemyLoc.y = loc.y;
		}
	}	
	
	FORALL(enemyControlCenters, building){
		ScalarPoint loc = building->second;
		distance = sqrt(square(loc.x - locCluster.x) + square(loc.y - locCluster.y));
		if(distance < bestDistance){
			bestDistance = distance;
			enemyLoc.x = loc.x;
			enemyLoc.y = loc.y;
		}
	}
	return enemyLoc;
}


void ClientEventHandler::buildNewBase(const Map<GameTile> &map){

	/* Obtenemos la unidad worker */
	GameObj* workerNB = newBase->removeWorker();
	
	/* Obtenemos las unidades militares */
	Vector<GameObj*> defenseNB = newBase->removeDefense();
	
	/* Obtenemos la localizacion */
	TerrainBase::Loc locNB = newBase->getLoc();
	
	/* Obtenemos la reserva de minerales */
	Finance::Ticket * reserve = newBase->getFunds();
	
	/* La nueva base se crea sin ningun edificio */
	Vector<GameObj*> vectorVacio;
	
	delete newBase;
	newBase = NULL;
	
	SpecialBaseCommander *bc = new SpecialBaseCommander(ms, vectorVacio, locNB, &controlCenters, &workers, &marines, &tanks);
	bases.push_back(bc);
	
	bc->createDefense(defenseNB);	
	bc->setAskingForBuilder(true);	
	bc->onReceiveUnit(workerNB, Watcher::NONE);	
	
	if (bc->canBuild("controlCenter") && fm.available() > fm.get_cost("controlCenter") ){	
		bc->buildBuilding("controlCenter", reserve);	
	}
	
		
	/* Obtenemos los clusters */
	Vector<boost::shared_ptr<MineralCluster> > clusters = simpleGather->getMineralClusters();
	
	/* Un cluster pertenece a una base si dista menos de 200 puntos
	 del control center y si el cluster es accesible */
	asignarClusters(clusters, map);
	
	simpleGather->comprobarWorkersGather();	
}


void ClientEventHandler::examineProblems(){
		
	size_t nWorkers = simpleGather->countWorkers() + workers.size() + scouting->getNumberScouts();
	
	/* El problema no se puede resolver, hay que cambiar la unidad worker y no se dispone de otra unidad */
	if((((newBase->retry()) || (newBase->getCambiarUnidad())) && (nWorkers <= 0)) || (newBase->retry() == false)){
		
		GameObj* workerNB = newBase->removeWorker();
		Vector<GameObj*> defenseNB = newBase->removeDefense();
		Finance::Ticket * funds = newBase->getFunds();

		delete newBase;
		newBase = NULL;

		size_t j = defenseNB.size();

		for(size_t i = 0; i < j; i++){
			GameObj* gob = defenseNB.back();
			defenseNB.pop_back();
			
			if(gob->bp_name() == "marine")
				marines.push_back(gob);
			else if(gob->bp_name() == "tank")
				tanks.push_back(gob);			
		}

		if(workerNB != NULL)
			workers.push_back(workerNB);

		funds->release();
		
		
	/* El problema se puede resolver, falta una unidad worker o se debe cambiar y se dispone de ella */
	}else if(((newBase->retry()) || (newBase->getCambiarUnidad())) && (nWorkers > 0)){
		
		GameObj* gob = newBase->removeWorker();				
		GameObj* unit = NULL;
				
		/* Obtenemos un worker que no tenia trabajo asignado*/
		if(workers.size() > 0){			
			unit = workers.back();
			workers.pop_back();		
			
		/* Obtenemos un worker que estaba realizando gather */
		}else if(simpleGather->countWorkers() > 0){			
			unit = simpleGather->removeWorkerGather();			
			
		/* Obtenemos un worker que estaba realizando scouting */
		}else if (scouting->getNumberScouts() > 0){			
			unit = scouting->removeScout();			
		}		
				
		assert(unit != NULL);				
		newBase->addNewWorker(unit);		
		
		if(gob != NULL)
			workers.push_back(gob);
						
	}	
}


void ClientEventHandler::evaluateAttack(SpecialDefenseCommander ** attack, bool weAreUnderAttack){

	Vector<sint4> posBases;	
	size_t nMarines, nTanks;
	float force;
	sint4 i = 0;
	bool cancelAttack = false;

	/* Un ataque se cancela solo si no esta atacando (es decir, si en su
	 radio de accion no hay ninguna unidad enemiga), el enemigo esta atacando
	 a una de las bases y en la base afectada hay un force < 10 */
	if(((*attack)->isUnderAttack() == false) && (weAreUnderAttack == true)){
	
		/* Hay que obtener las bases que nos estan atacando y
		 la cantidad de militares que hay en cada una */
		FORALL(bases, base){
		
			/* Nos estan atacando */
			if(((*base)->isDefenderOperative()) && ((*base)->isBaseUnderAttack() == true)){
				nMarines = (*base)->countUnitsDefender("marine");
				nTanks = (*base)->countUnitsDefender("tank");
				force = nMarines + (nTanks * 3);
				if(force < 10){
					posBases.push_back(i);					
					cancelAttack = true;
				}
			}
			i++;
		}
		
		
		/* Hay que cancelar el ataque */
		if(cancelAttack == true){
									
			/* Se obtienen todas las unidades del ataque */
			Vector<GameObj*> vector = (*attack)->getAllUnits();
			
			delete (*attack);
			(*attack) = NULL;			
			
			/* Solo necesita ayuda una base */
			if(posBases.size() == 1){
											
				sint4 pos = posBases.back();
				
				/* La defensa de la base esta operativa */
				if(bases[pos]->isDefenderOperative()){
				
					FORALL(vector, gob){
						bases[pos]->addUnitToDefender(*gob);
					}
				
				/* La defensa de la base no esta operativa */
				}else{
					bases[pos]->createDefense(vector);					
				}
							
			/* Necesitan ayuda las dos bases */
			}else if(posBases.size() == 2){
						
				size_t nMarinesB1 = 0;
				size_t nMarinesB2 = 0;
				size_t nTanksB1 = 0;
				size_t nTanksB2 = 0;
				
				FORALL(vector, gob){
				
					/* La unidad es un marine */
					if((*gob)->bp_name() == "marine"){
					
						/* Unidad para la base 1 */
						if(nMarinesB1 < nMarinesB2){							
							if(bases[0]->isDefenderOperative()){
								nMarinesB1++;
								bases[0]->addUnitToDefender(*gob);
							}							
						
						/* Unidad para la base 2 */
						}else{
							if(bases[1]->isDefenderOperative()){
								nMarinesB2++;
								bases[1]->addUnitToDefender(*gob);
							}							
						}					
					
					/* La unidad es un tank */
					}else if((*gob)->bp_name() == "tank"){
					
						/* Unidad para la base 1 */
						if(nTanksB1 < nTanksB2){
							
							if(bases[0]->isDefenderOperative()){
								nTanksB1++;
								bases[0]->addUnitToDefender(*gob);
							}							
							
						/* Unidad para la base 2 */
						}else{
							if(bases[1]->isDefenderOperative()){
								nTanksB2++;
								bases[1]->addUnitToDefender(*gob);
							}							
						}
					
					}									
				}
			
			}else{
				assert("ERROR. Solo se contemplan dos bases en el juego");
			}
			vector.clear();			
		}		
	}
}


void ClientEventHandler::createAttack(std::string tipo){

	/* Un ataque de tipo low significa utilizar las unidades militares de
	 la base que menos force disponga. En el caso de que solo haya una
	 base, se utilizaran un tercio de las unidades militares disponibles. */
	
	/* Un ataque de tipo medium signigica utilizar las unidades militares
	 de la base que mayor force disponga. En el caso de que solo haya una
	 base, se utilizaran dos tercios de las unidades militares disponibles. */
	 
	/* Un ataque de tipo high signigica utilizar todas las unidades militares disponibles */
		
	/* Si se dispone de una base */
	if(bases.size() == 1){
	
		TerrainBase::Loc locEnemy = getLocationEnemyToNewBase(bases[0]->getCentre());
		Vector<GameObj*> units;
				
		if(tipo == "low")	
			units = bases[0]->getUnitsToNewBase(1);
		else if(tipo == "medium")	
			units = bases[0]->getUnitsToNewBase(2);
		else if(tipo == "high")
			units = bases[0]->getUnitsToNewBase(3);
		
		
		if(attack1 == NULL){
			attack1 = new SpecialDefenseCommander(ms, locEnemy, 128, 0, false);
			FORALL(units, unit)
				attack1->addUnit(*unit);
				
		}else if(attack2 == NULL){
			attack2 = new SpecialDefenseCommander(ms, locEnemy, 128, 0, false);
			FORALL(units, unit)
				attack2->addUnit(*unit);
			
		}		
			
	
	/* Si se dispone de bos bases */
	}else if(bases.size() == 2){
	
		size_t forceB0 = 0;
		size_t forceB1 = 0;
		size_t nMarines, nTanks;
		Vector<GameObj*> unitsToAttack;
		Vector<GameObj*> unitsToDefense;
		TerrainBase::Loc locEnemy;
		sint4 posDefense;
		
		nMarines = bases[0]->countUnitsDefender("marine");
		nTanks = bases[0]->countUnitsDefender("tank");
		forceB0 = nMarines + (nTanks * 3);
		
		nMarines = bases[1]->countUnitsDefender("marine");
		nTanks = bases[1]->countUnitsDefender("tank");
		forceB1 = nMarines + (nTanks * 3);
	
		/* Ataque con preucacion */
		if(tipo == "low"){
									
			/* La base 0 tiene menos force */
			if(forceB0 < forceB1){			
				locEnemy = getLocationEnemyToNewBase(bases[0]->getCentre());
				unitsToAttack = bases[0]->getUnitsToNewBase(3);
				unitsToDefense = bases[1]->getUnitsToNewBase(1.5);				
				posDefense = 0;
				
			}else{
				locEnemy = getLocationEnemyToNewBase(bases[1]->getCentre());
				unitsToAttack = bases[1]->getUnitsToNewBase(3);
				unitsToDefense = bases[0]->getUnitsToNewBase(1.5);
				posDefense = 1;
			}						
					
		/* Ataque moderado */
		}else if(tipo == "medium"){
				
			/* La base 0 tiene mayor force */
			if(forceB0 > forceB1){			
				locEnemy = getLocationEnemyToNewBase(bases[0]->getCentre());
				unitsToAttack = bases[0]->getUnitsToNewBase(3);
				unitsToDefense = bases[1]->getUnitsToNewBase(1.5);
				posDefense = 0;			
				
			}else{
				locEnemy = getLocationEnemyToNewBase(bases[1]->getCentre());
				unitsToAttack = bases[1]->getUnitsToNewBase(3);
				unitsToDefense = bases[0]->getUnitsToNewBase(1.5);
				posDefense = 1;
			}
					
		
		/* Ataque campal */
		}else if(tipo == "high"){
		
			unitsToAttack = bases[0]->getUnitsToNewBase(3);
			unitsToDefense = bases[1]->getUnitsToNewBase(3);
									
			FORALL(unitsToDefense, unit)
				unitsToAttack.push_back(*unit);
				
			unitsToDefense.clear();
									
			if(forceB0 > forceB1)
				locEnemy = getLocationEnemyToNewBase(bases[0]->getCentre());				
			else
				locEnemy = getLocationEnemyToNewBase(bases[1]->getCentre());
				
			posDefense = -1;
		}	
		
		
		if(attack1 == NULL){
			attack1 = new SpecialDefenseCommander(ms, locEnemy, 128, 0, false);
			FORALL(unitsToAttack, unit)
				attack1->addUnit(*unit);
			
		}else if(attack2 == NULL){
			attack2 = new SpecialDefenseCommander(ms, locEnemy, 128, 0, false);
			FORALL(unitsToAttack, unit)
				attack2->addUnit(*unit);
			
		}		
		
		
		if(posDefense != -1){
			FORALL(unitsToDefense, unit)
				bases[posDefense]->addUnitToDefender(*unit);		
		}	
	
	}		
}


sint4 ClientEventHandler::getPosBase(TerrainBase::Loc locCC){
	sint4 posBase = -1;
	
	/* Solo hay una base */
	if(bases.size() == 1){
		posBase = 0;
	
	/* Hay dos bases */
	}else if(bases.size() == 2){
		
		float bestDistance = 1448;
		float distance;
		sint4 i = 0;
				
		FORALL(bases, base){
			TerrainBase::Loc locBase;
			locBase = (*base)->getCentre();
			distance = sqrt(square(locCC.x - locBase.x) + square(locCC.y - locBase.y));
			if(distance < bestDistance){
				bestDistance = distance;
				posBase = i;
			}
			i++;
		}
		
	/* Caso de error */
	}else{
		assert(bases.size() <= 2);
		assert(bases.size() >= 1);
	}	
	
	assert(posBase != -1);
	return posBase;
}


