#ifndef SpecialBaseCommander_H
#define SpecialBaseCommander_H

/* 
 * Base Commander
 * Anthony Marcar
*/

#include "SpecialCommander.H"
#include "ClientModule.H"
#include "TerrainBase.H"
#include "Finance.H"
#include "SimpleGather.H"

class GameStateModule;
class GameObj;
class SpecialBuildCommander;
class SpecialDefenseCommander;

class SpecialBaseCommander : public SpecialCommander, ClientModule{
    
	struct BuildTask{
        std::string       bp_name;
        Finance::Ticket * funds;
        
        BuildTask(const std::string & bp_name, Finance::Ticket * funds) : bp_name(bp_name), funds(funds) {}
    };
	
public:

    class Building{
	
        ModuleSet &       ms;
        SpecialBaseCommander *   base;
        GameObj *         building;
        std::string       training;
        std::string       action;
        Finance::Ticket * funds;
        Watcher::WatchID  watch;
        bool              confirmed;
        bool              complete;
		
    public:
        Building(ModuleSet & ms, SpecialBaseCommander * base, GameObj * building);
        ~Building();
        GameObj * getBuilding();
        bool isBusy() const;
        const std::string & getTrainingBP() const;
        bool canTrain(std::string bp_name) const;
        void train(std::string bp_name, Finance::Ticket * funds);
        Watcher::WatchID getWatch() const;
        void update();
    };
	
private:
    TerrainBase::Loc centreLoc;
	sint4 radius;

    GameObj *          controlCenter;
	SpecialDefenseCommander * defender;
	
	
	Vector<GameObj*>* workers;
	Vector<GameObj*>* marines;
	Vector<GameObj*>* tanks;
	Vector<GameObj*>* controlCenters;	

    std::set<SpecialBuildCommander*> activeBuilders;
    std::set<SpecialBuildCommander*> finishedBuilders;

    std::map<std::string, size_t> unitsQueued; ///< Number of units queued
    std::map<GameObj*, Building*> obj2building; ///< Job information
    
	Vector<GameObj*> builders;
    bool askingForBuilder;     ///< Whether or not we currently need builder

    std::queue<BuildTask> unitsToBuild;     ///< FIFO queue of training tasks
    std::queue<BuildTask> buildingsToBuild; ///< FIFO queue of build tasks


    //////////////////////////Methods////////////////////////////////

    ///initialises the Control Center and creates a gather commander
    void init(const Vector<GameObj*> & initialUnits);
    ///tells buildings to train any units in the queue
    void buildUnitsInQueue();

    void buildBuildingsInQueue();
    
    /**
     * where everything is executed
     */
    void exec();

public:

	Vector<boost::shared_ptr<MineralCluster> > clusters;

    //*************************Constructor******************************************

    /**
     * creates a new base commander
     * @param units the units under this commanders control (includes buildings)   
     */
    SpecialBaseCommander(ModuleSet& ms, Vector<GameObj*> units, TerrainBase::Loc centre,
						Vector<GameObj*>* vectorCC, Vector<GameObj*>* vectorWorkers,
						Vector<GameObj*>* vectorMarines, Vector<GameObj*>* vectorTanks, sint4 updateTime_=25);
    
    ~SpecialBaseCommander();

    void buildBuilding(const std::string & buildingBPname, Finance::Ticket * funds);

    /**
     * adds units to be built to the queue
     * @param unitBPname the bp name of the unit to train
     */
    void buildUnit(const std::string & unitBPname, Finance::Ticket * funds);

    /**
     * returns whether this commander can build the specified building
     * @param objectBPname the object to query (unit or building)
     * @return whether this unit can be built
     */
    bool canBuild(const std::string & objectBPname);

    TerrainBase::Loc getCentre(){
        return centreLoc;
    }
	
	void setNewCentre(TerrainBase::Loc centerGob);
	
	void setNewRadius();

    void onReceiveUnit( GameObj * unit, Watcher::WatchID );
	
    void onLoseUnit( GameObj * unit );

    /**
     * changes flags and sets vectors for the execution cycle.
     */
    bool onEvent(const Event &event);
    
    size_t countQueued(const std::string & bp_name) const;   
	
	void setAskingForBuilder(bool value);
	
	bool isDefenderOperative();
	
	void addUnitToDefender(GameObj* gob);
	
	size_t countUnitsDefender(const std::string & bp_name);
	
	bool isBaseUnderControl();
	
	bool isBaseUnderAttack();
	
	size_t countBuilding();

	sint4 getLastAttack();
	
	float isThereBuilding(std::string bp_name);
	
	float buildingNow(std::string bp_name);
	
	TerrainBase::Loc getLocationControlCenter();
	
	size_t countBuildings(const std::string & bp_name);
	
	GameObj* getControlCenter();
	
	void createDefense(Vector<GameObj*> units);
	
	Vector<GameObj*> getUnitsToNewBase(float a);
};

#endif
