#include "SpecialBuildCommander.H"
#include "GameObj.H"

#include "TerrainAnalysis.H"
#include "Delegate.H"

const sint4 SpecialBuildCommander::FROM     = EventFactory::new_who();
const sint4 SpecialBuildCommander::FINISHED = EventFactory::new_what();
const sint4 SpecialBuildCommander::FAILED   = EventFactory::new_what();

SpecialBuildCommander::SpecialBuildCommander(SpecialCommander *         mySuperior, 
                               GameObj *           worker,
                               const std::string & bp_name,
							   TerrainBase::Loc    nearLoc,
                               Finance::Ticket *   funds,
							   sint4               updateTime) :
    SpecialCommander(mySuperior, updateTime),
    tam(moduleSet.get_terrain_analysis()),
    wm(moduleSet.get_watcher()),
    mc(moduleSet.get_movement(), "Default", "Default"),
    bp_name(bp_name),
	nearLoc(nearLoc),
    funds(NULL),
    worker(NULL),
    building(NULL),
    state(PLANNING),
    tryingToBuild(false),
    buildWatch(Watcher::NONE),
    completeWatch(Watcher::NONE)
{
    mc.setStatusProcedure(make_delegate(this, &SpecialBuildCommander::handleMovementStatus));

	// If a finance ticket was given, ensure that everything is in
	// order. Otherwise, we will simply have to try to build, and
	// see if any funding is available.
	if(funds){
		// Take the funds that we need
		uint4 amount = moduleSet.get_finance().get_cost(bp_name);
		assert( funds->available() >= amount &&
            "Not enough funds to build requested building" );
		funds->trim(amount);
	}
    this->funds = funds;

    addUnit( worker );
}

SpecialBuildCommander::~SpecialBuildCommander(){
    // If we failed before spending our funds
    if(funds != NULL){
        // Release them so someone else can use them
        funds->release();
    }
}

void SpecialBuildCommander::handleMovementStatus(const Movement::Task & task){
    
	if(state == ENROUTE){
		
		switch(task.getStatus()){
		
		case Movement::Task::SUCCESS:
			state         = READY;
			tryingToBuild = false;			
			break;
        case Movement::Task::PLAN_FAILURE:
		case Movement::Task::MOVE_FAILURE:
			worker->get_center(nearLoc.x, nearLoc.y);
			state = PLANNING;						
			break;
		default: break;
		}
    }
}

void SpecialBuildCommander::onFail(){

    // Clean up any watcher notifications
    if(state == READY && tryingToBuild){
        wm.cancelNotify(buildWatch);
    }
	
    if(state == BUILDING){
        wm.cancelNotify(completeWatch);
    }
    
    state = STATE_FAILED;
}

void SpecialBuildCommander::onReceiveUnit(GameObj * unit, Watcher::WatchID watch){

    if(unit->bp_name() == "worker" || unit->bp_name() == "dactyl" || unit->bp_name() == "vector"){
        worker = unit;
        state = PLANNING;        
    }

    if(state == READY && watch == buildWatch){
        building      = unit;
        tryingToBuild = false;
        state         = BUILDING;

        sint4 x, y;
        building->get_center(x, y);

        sint4 player = moduleSet.get_game_state().get_game().get_client_player();
        completeWatch = wm.notifyOnCondition(this,
            Watcher::And(Watcher::InCircle(x, y, 16),
                         Watcher::TypeIs(bp_name),
                         Watcher::OwnerIs(player),
                         Watcher::BuildingDone()),
            Watcher::TotalCountIs(Watcher::GREATER_THAN, 0),
            8, // Check once per second
            true); // Autodelete

		if(funds){
			// Since the building is confirmed, we can spend our funds now
			funds->spend();
			funds = 0;
		}        
    }
}

void SpecialBuildCommander::onLoseUnit(GameObj * unit){
    
	if( unit == worker ){ // Handle worker death
    
        worker = NULL;
        if( state != STATE_FAILED ){
            myCommander->handle_event(BuildFailedEvent(this, NULL, BuildFailedEvent::WORKER_KILLED));
            onFail();
        }
		
    }else if( unit == building ){ // Handle building destruction
    
        building = NULL;
        if( state != STATE_FAILED ){
            myCommander->handle_event(BuildFailedEvent(this, NULL, BuildFailedEvent::BUILDING_DESTROYED));
            onFail();
        }
    }
}

bool SpecialBuildCommander::onEvent(const Event &event){

    if(worker == NULL){
        return false;
    }

    if(state == BUILDING){

        if(event.get_who() == Watcher::WatcherModule::FROM){

            if(event.get_what() == Watcher::WatcherModule::CONDITION){

                const Watcher::ConditionEvent & e =
                    reinterpret_cast<const Watcher::ConditionEvent &>(event);

                if(e.getID() == completeWatch){                    
                    state = STATE_FINISHED;
                    myCommander->handle_event(BuildFinishedEvent(this, worker, building));
                    state = STATE_FINISHED;
                    return true;
                }
            }
        }
    }

    return false;
}

void SpecialBuildCommander::exec(){

    if(worker == NULL){
        return;
    }

    switch(state){
        case PLANNING:
            doPlanning();
            break;
        case ENROUTE:
            break;
        case READY:
            tryToBuild();
            break;
        case BUILDING:
        case STATE_FINISHED:
        case STATE_FAILED:
            break;
    }
}

void SpecialBuildCommander::doPlanning(){
    assert(state == PLANNING);

    findBuildSite();
    moveToNextSite();
}

void SpecialBuildCommander::moveToNextSite(){
    
	if(nextSite < buildSites.size()){
		const TerrainBase::Loc & site = buildSites[nextSite++];

		assert(site.x >= 0);
		assert(site.x <= 1023);
		assert(site.y >= 0);
		assert(site.y <= 1023);

		mc.moveUnit(worker, Movement::TouchPoint(site));

        state = ENROUTE;       
    
	}else{
                
        /* Continue to plan until space becomes available */
        /*myCommander->handle_event( 
          BuildFailedEvent(this, worker, BuildFailedEvent::NO_BUILD_SITE));*/
        state = PLANNING;
    }
}

void SpecialBuildCommander::tryToBuild(){
    assert(worker != NULL);

    std::string buildAction = "build_";
    buildAction += bp_name;

    sint4 x,y;
    worker->get_center(x,y);

    if(tryingToBuild){
        countDown--;
        if(countDown <= 0){
 
            wm.cancelNotify(buildWatch);
            tryingToBuild = false;
            moveToNextSite();
        }
    
	}else{
        sint4 player = moduleSet.get_game_state().get_game().get_client_player();
        buildWatch = wm.notifyOnCreation(this,
             Watcher::And(Watcher::OwnerIs(player), 
                          Watcher::TypeIs(bp_name),
                          Watcher::InCircle(x, y, 16)),
             1000,  // Very high priority
             true); // Autodelete
        tryingToBuild = true;
        countDown = 3;

        Vector<sint4> args;
        args.push_back(x);
        args.push_back(y);
        worker->set_action(buildAction, args);
       
    }
}

#include "GridAbstraction.H"
using TerrainAnalysis::GridAbstraction;

bool SpecialBuildCommander::findBuildSite(){

    assert( worker != NULL );

    sint4 distance = 320;
    GridAbstraction ga(moduleSet, nearLoc.x, nearLoc.y, distance);	

    ga.set_object_border("mineral", 16);
    ga.set_object_border("controlCenter", 16);
    ga.set_object_border("barracks", 16);
    ga.set_object_border("factory", 16);

    sint4 height = ga.sample_height(nearLoc);
    ga.set_passable_height(height);

    buildSites.clear();
    nextSite = 0;
    ga.place_building( bp_name, nearLoc, buildSites );
    return !buildSites.empty();
}
