#include "SpecialCommander.H"

/**
 * Decriments updateCounter by 1.
 * Recursively calls update on all Commanders.
 * When updateCounter == 0 updateCounter is set to 
 * a new random number (maxUpdateCount = maximum) and exec() is called.
 * @see updateCounter
 * @see maxUpdateCount
 * @see exec()
 */
void SpecialCommander::update(){
    if (updateCounter==0){
        exec();
        updateCounter=rand()%maxUpdateCount+1;
    }
	
    FORALL(comms, c){
        (*c)->update();		
    }
	
    updateCounter--;
}

    
/**
 * Constructor for Commander.
 * @param moduleSet reference to the global module set.
 * @param maxUpdate an integer that will be used as the highest possible random number of update() calls before exec() is run.
 * @see maxUpdateCount
 * @see updateCounter
 * @see ~Commander()
 */  
SpecialCommander::SpecialCommander(ModuleSet& moduleSet, int maxUpdate) :
    constructed(true), myCommander(NULL), moduleSet(moduleSet), maxUpdateCount(maxUpdate){ 
    
	srand((unsigned)time(0));
	updateCounter=rand()%maxUpdateCount+1;
	commanderID=EventFactory::new_who();   
}


/**
 * Constructor for Commander.
 * @param mySuperior a pointer of type Commander to the Commander's superior.
 * @see myCommander
 * @param maxUpdate an integer that will be used as the highest possible random number of update() calls before exec() is run.
 * @see maxUpdateCount
 * @see updateCounter
 * @see ~Commander()
 */  
SpecialCommander::SpecialCommander(SpecialCommander* superior, int maxUpdate) :
    constructed(true), myCommander(superior), moduleSet(superior->moduleSet), maxUpdateCount(maxUpdate){ 
    
	srand((unsigned)time(0));		
	updateCounter=rand()%maxUpdateCount+1;
	superior->comms.insert(this);
	commanderID=EventFactory::new_who();
}

/**
 * Destroys the Commander and all subcommanders recursively.
 * Sends units to superior Commander.
 * @see commVector
 * @see myCommander
 * @see unitVector
 * @see Commander()
 */
SpecialCommander::~SpecialCommander(){
    constructed = false;   
	
	FORALL(comms, c){
        delete (*c);
    }
    comms.clear();

    if(myCommander != NULL){
        while(!units.empty()){
            std::set<GameObj*>::iterator it = units.begin();
            GameObj * obj = (*it);
            removeUnit( obj );
            myCommander->addUnit( obj );
        }  
		myCommander->comms.erase(this);     
    }
}
    
bool SpecialCommander::handle_event(const Event &event){
    
	if( event.get_who() == Watcher::WatcherModule::FROM ){
        
		if( event.get_what() == Watcher::WatcherModule::CREATION ){
            
			const Watcher::CreationEvent & e = reinterpret_cast<const Watcher::CreationEvent&>(event);
            const std::set<GameObj*> nUnits = e.getUnits();
            std::set<GameObj*>::const_iterator it;
            
			for(it = nUnits.begin(); it != nUnits.end(); it++){
                
				if(((*it)->bp_name() == "controlCenter") || ((*it)->bp_name() == "barracks") || ((*it)->bp_name() == "factory")){
					moduleSet.get_watcher().notifyOnDeath( this, (*it) );
					units.insert((*it));					
				}
				
                if(constructed){					
                    onReceiveUnit((*it), e.getID());
                }
            }
        
		}else if( event.get_what() == Watcher::WatcherModule::DEATH ){
            
			const Watcher::DeathEvent & e = reinterpret_cast<const Watcher::DeathEvent&>(event);
            const std::set<GameObj*> dUnits = e.getUnits();
            std::set<GameObj*>::const_iterator it;
            
			for(it = dUnits.begin(); it != dUnits.end(); it++){
                if(constructed){
                    onLoseUnit((*it));
                }
                units.erase((*it));
            }
        }
    }

    return onEvent(event);
}

sint4 SpecialCommander::getID() const{
    return commanderID;
}


void SpecialCommander::addUnit(GameObj * unit){
    
	if((unit->bp_name() == "controlCenter") || (unit->bp_name() == "barracks") || (unit->bp_name() == "factory")){
		moduleSet.get_watcher().notifyOnDeath(this, unit);
		units.insert(unit);		
	}
	
    if(constructed){
        onReceiveUnit(unit, Watcher::NONE);
    }
}

void SpecialCommander::addUnit2(GameObj * unit){	
	units.insert(unit);	    
}

void SpecialCommander::removeUnit(GameObj * unit){
    if(constructed){
        onLoseUnit(unit);
    }
    units.erase(unit);
	moduleSet.get_watcher().cancelNotifyOnDeath( this, unit );
}

size_t SpecialCommander::countOwned(const std::string & bp_name) const{
    size_t count = 0;
    
	if(bp_name != ""){
		FORALL(units, u){
			if((*u)->bp_name() == bp_name){
				count++;
			}
		}
	
	}else{
		count += units.size();
	}
	
    return count;
}
