#ifndef SpecialRoute_H
#define SpecialRoute_H

#include "TerrainBase.H"
#include "History.H"
#include "MineralCluster.H"

class SimpleGather;
class MineralCluster;
class GameObj;

/**
 * Routes contain information that a worker needs to be able to move back and forth
 * between a control centre and a mineral.  To move, workers actually call the route
 * which has functions moveToMineral() and moveToCC().  This is for the case that 
 * different styles of routes could eventually exist.  A route can have multiple 
 * workers assigned to it depending upon its maxWorkers which is based on distance
 * between the control center and access point.
 */ 
class SpecialRoute{

protected:
    /// the commander who created this route (for pathfinding purposes)
    SimpleGather *commander;

    /// the mineral associated with this route
    const History::Record *mineralRec;

    TerrainBase::Loc accessPoint;

    /// the control center associated with this route
    GameObj *controlCenter;

    /// the id of this route (unique)
    sint4 id;

    /// the number of workers currently using this route
    sint4 currentWorkers;

    /// the maximum number of workers who can use this route
    sint4 maxWorkers;

    /// The distance of the route
    real8 routeLength;

    /// the number of path failures that have occurred
    sint4 planFailures;

public:
    /**
     * creates a route
     * @param commander the commander who created this route (for pathfind commands)
     * @param mineralRec the mineral (history record) associated with this route
     * @param accessPoint the accessPoint for the mineral
     * @param cc the control center associated with this route
     * @param max the maximum number of workers that can work this route
     */
    SpecialRoute(SimpleGather *commander,
          const History::Record *mineralRec,
          TerrainBase::Loc accessPoint,
          GameObj *cc,
          sint4 max=1);
	

    virtual ~SpecialRoute();

    /**
     * increments the number of workers on this route
     */
    void addWorker();

    /**
     * returns the access point of this route
     */
    TerrainBase::Loc getAccessPoint() const;

    /**
     * returns the control center associated with this route
     */
    GameObj* getCC() const;

    /**
     * returns the number of workers assigned to this route
     */
    sint4 getCurrentWorkers() const;

    /**
     * returns the id of this route
     */
    sint4 getID() const;

    /**
     * returns the maximum number of workers that can be assigned to this route
     */
    sint4 getMaxWorkers() const;

    /**
     * returns the history record of the mineral associated with this route
     */
    const History::Record* getMineralRec() const;

    /**
     * Returns the real length of this route based on path finding rather than dist
     */
    real8 getRealRouteLength() const;

    /**
     * returns the length of this route
     */
    real8 getRouteLength() const;

    /**
     * returns the straight distance between the control center and access point
     */
    real8 getSimpleRouteLength() const;

    /**
     * returns whether the history record is active (still exists)
     */
    bool mineralIsActive() const;

    /**
     * calls the commander to move the worker to its mineral
     */
    void moveToMineral(GameObj *worker) const;

    /**
     * calls the commander to move the worker to its control center
     */
    void moveToCC(GameObj *worker) const;

    /**
     * decrements the number of workers on this route
     */
    void removeWorker();

    /**
     * sets the id of this route
     */
    void setID(sint4 id_);

    /**
     * sets the length of this path
     */
    void setRouteLength(real8 length);

    /// returns the number of path failures for this route
    sint4 getPlanFailures() { return planFailures; }
    
    /// increments the number of path failures for this route
    void planFailureOccurred();
    
    /// decrements the number of path failures for this route
    void planSuccessOccurred();
};

#endif
